#%% VIF (Variance Inflation Factor)

import numpy as np
import statsmodels.api as sm
import RG as rg

def vif(X0):

    # Make a bunch of models
    # Each model contains all but 1 variable
    # That variable is predicted using the others
    # This is acheived with the "selector" variable
    n,k = X0.shape
    R2_i = np.zeros((k,1))
    VIF_i = np.zeros((k,1))
    
    for i in range(k):
        selector = [X0 for X0 in range(X0.shape[1]) if X0 != i]
        X = X0[:,selector]
        y = X0[:,i]
    
        X = sm.add_constant(X) ## Add an intercept (beta_0) to our model
        b, b_int, model, R2, F, p = rg.regress(y, X)
        R2_i[i] = R2
        VIF_i[i] = 1/(1-R2)
        
    return VIF_i, R2_i
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun  2 21:44:58 2020

@author: ryangosselin
"""

# LDA: linear discriminate analysis

import numpy as np

def lda(nbPC, *args):
    ''' 
    
    LDA: linear discriminate analysis
    [T,P,SSX] = LDA(nbPC, X1, X2, Xi, ...)
    
    INPUT
    npPC [1 x 1] <numpy.ndarray>
        number of components in the LDA decomposition
    Xi [ni x k] <numpy.ndarray>
        spectra
        ni samples
        k variables
    
    OUTPUT
    T [(n1+n2+...) x nbPC] 
        scores
    P [nbPC x k] 
        loadings
    SSX [1 x nbPC] 
        Sum of squared variance explained by each component
        
    '''

    # Tuple containing n block matrices
    varargin = args
    n_block = len(varargin)
    
    # Number of columns in the data (dimensionality)
    k = varargin[0].shape[1]

    # Number of observation in each block
    N = np.zeros((n_block,1))
    for i in range(n_block):
        ni = varargin[i].shape[0]
        N[i] = ni

    # Means
    MU = np.zeros((n_block,k))
    for i in range(n_block):
        xi = varargin[i]
        mu = np.mean(xi,axis=0)
        MU[i,:] = mu
    
    # Overall mean
    MU_overall = np.mean(MU,axis=0)
    MU_overall = MU_overall[np.newaxis].T
    
    # Within-class scatter matrix
    Sw = np.zeros((k,k))
    for i in range(n_block):
        si = np.cov(varargin[i].T) # Class covariance matrices
        Sw = Sw + si
    
    # Between-class statter matrix
    Sb = np.zeros((k,k)) 
    for i in range(n_block):
        ni = N[i]
        mui = MU[i,:][np.newaxis].T
        sbi = ni * (mui-MU_overall)@(mui-MU_overall).T
        Sb = Sb + sbi
    Sb = Sb / np.sum(N)

    # Compute LDA projections
    Swinv = np.linalg.pinv(Sw) #Use pseudo-inverse to avoir rank issues
    SwinvSb = Swinv@Sb
    
    U,D,V = np.linalg.svd(SwinvSb)
    P = U[:,:nbPC]
    
    # Correct for rotational ambiguity
    for i in range(P.shape[1]):
        if P[:,i][np.newaxis]@np.ones((P.shape[0],1)) < 0:
            P[:,i] = -P[:,i]
    
    # Create overall X matrix containing n blocks    
    X = []
    for i in range(n_block):
        xi = varargin[i]
        if (i==0):
            X = xi
        else:
            X = np.append(X,xi,axis=0)
    
    T = X@P

    # Sum of squares -----------
    SSX = np.zeros((T.shape[1],1)) 
    for i in range(T.shape[1]):
        t = T[:,i][np.newaxis].T
        p = P[:,i][np.newaxis]
        Xhat = t@p
        #Xhat = T[:,i][np.newaxis].T@P[:,i][np.newaxis].T
        ssX0 = np.sum(np.sum(X * X)) # Each element squared
        ssX = np.sum(np.sum(Xhat * Xhat))
        ssX = ssX / ssX0
        SSX[i] = ssX
    # --------------------------

    
    # Sort by SSX   
    sort = np.argsort(SSX.T) 
    sort = np.fliplr(sort) # Decreasing order
    SSX = SSX[sort]
    T = np.squeeze(T[:,sort])
    P = np.squeeze(P[:,sort])

    return T,P,SSX


# PLS: partial least squares

import numpy as np

def pls(X, Y, nbPC):
    ''' 
    
    PLS: partial least squares
    [T, U, P, W, Q, B, Yhat, SSX, SSY] = PLS(X, Y, nbPC)
    
    INPUT
    X [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    Y [n x m] <numpy.ndarray>
        responses
        n samples
        m variables
    npPC [1 x 1] <numpy.ndarray>
        number of components in the PLS decomposition
    
    OUTPUT
    T [n x nbPC] 
        X-block scores
    U [n x nbPC] 
        Y-block scores           
    P [k x nbPC] 
        X-block loadings
    W [k x nbPC] 
        X-block weights        
    Q [m x nbPC] 
        Y-block weights         
    B [nbPC x nbPC]
        Y-block weights 
    Yhat [n x nbPC]
        Y predictions        
    SSX [nbPC x 1] 
        X-block sum of squared variance explained by each component
    SSY [nbPC x 1] 
        Y-block sum of squared variance explained by each component
        
    '''
    

    s0, s1 = X.shape
    s0, s2 = Y.shape
    T = np.zeros(shape=(s0, nbPC))
    U = np.zeros(shape=(s0, nbPC))
    P = np.zeros(shape=(s1, nbPC))
    W = np.zeros(shape=(s1, nbPC))
    Q = np.zeros(shape=(s2, nbPC))
    B = np.zeros(shape=(nbPC, nbPC))
    SSX = np.zeros(shape=(nbPC, 1))
    SSY = np.zeros(shape=(nbPC, 1))
    X0 = X # Save a copy
    Y0 = Y # Save a copy
    
    for i in range(0,nbPC):
        error = 1
        u_temp = np.random.randn(s0, 1)
        u = np.random.randn(s0, 1)
        
        while error > 1E-10:
            w = (X.T @ u) / np.linalg.norm(X.T @ u)
            
            # Fix rotational ambiguity
            onez = np.ones([len(w),1]) # Use as a reference
            if w.T@onez < 0:
                w = -w
                
            t = X @ w    
            q = (Y.T @ t) / np.linalg.norm(Y.T @ t)
            u = Y @ q
            # Check t convergence 
            error = sum(np.power((u-u_temp),2),0) # Squared error
            u_temp = u
        
        # Deflation
        p = (X.T@t)/(t.T@t)
        b = (u.T@t)/(t.T@t) # U = fct(T)
        X = X - t@p.T
        Y = Y - b*t@q.T
               
        # Save        
        W[:,i] = np.squeeze(w)
        P[:,i] = np.squeeze(p)
        T[:,i] = np.squeeze(t)
        U[:,i] = np.squeeze(u)
        Q[:,i] = np.squeeze(q)
        B[i,i] = b

        # Sum of squares X
        Xhat = t@p.T
        ssX0 = np.sum(np.sum(X0 * X0)) # Each element squared
        ssX = np.sum(np.sum(Xhat * Xhat))
        ssX = ssX / ssX0
        SSX[i] = ssX
 
        # Sum of squares Y 
        Yhat = b*t@q.T
        ssY0 = np.sum(np.sum(Y0 * Y0)) # Each element squared
        ssY = np.sum(np.sum(Yhat * Yhat))
        ssY = ssY / ssY0
        SSY[i] = ssY
        
    Yhat = T@B@Q.T # NIPALS
        
    return T, U, P, W, Q, B, Yhat, SSX, SSY



# Binary surface plot 3D
import numpy as np
import matplotlib.pyplot as plt

def binaire3D(i,j,beta,*args): # creates a model with beta coefficients : [i,j,ij,i2,j2]

    # Make mesh grid
    ii = np.arange(min(i),max(i),(max(i)-min(i))/100)
    jj = np.arange(min(j),max(j),(max(j)-min(j))/100)
    ii, jj = np.meshgrid(ii, jj) 
    ij2 = ii*jj
    ii2 = ii*ii
    jj2 = jj*jj
    
    yy = beta[0]*np.ones((ii.shape)) + beta[1]*ii + beta[2]*jj + \
    beta[3]*ij2 + beta[4]*ii2 + beta[5]*jj2
        
    fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
    ax.plot_surface(ii, jj, yy, cmap=plt.cm.plasma,linewidth=0, antialiased=False, alpha = 0.5)
    # ax.set_xlabel('a', fontsize=16)
    # ax.set_ylabel('b', fontsize=20)
    # ax.set_zlabel('y', fontsize=20)
    
    try : ax.set_xlabel(args[0],fontsize=16)
    except : ax.set_xlabel('$x_i$', fontsize=16)
    try : ax.set_ylabel(args[1],fontsize=16)
    except : ax.set_ylabel('$x_j$', fontsize=16)
    try : ax.set_zlabel(args[2], fontsize=16) 
    except : ax.set_zlabel('$y$', fontsize=16)    
        
    plt.show()




# MLR 
# Ryan Gosselin
# Version 2023


import numpy as np
import statsmodels.api as sm
import pandas as pd

def mlr(y, X):
    ''' 
    
    regress: MLR regress
    [b, b_int, R2, F, p] = mlr(y,X)
    
    INPUT
    X [n x k] <numpy.ndarray>
        independent variable
        n samples
        k variables
    y [n x 1] <numpy.ndarray>
        dependent variable
    
    OUTPUT
    b [k x 1]
        regression coefficients
    b_int [k x 2]
        95% confidence interval on regression coefficients  
    R2 [1 x 1]
        Coefficient of determination
    F [1 x 1]
        Fisher test: F value for the MLR model
    p [1 x 1]
        Fisher test: p value for the MLR model
    
    ''' 
    
    model = sm.OLS(y, X).fit()
    b = model.params
    b_int = model.conf_int(0.05)
    b = (b[np.newaxis]).T # Vector format for output
    
    R2 = model.rsquared
    F = model.fvalue
    p = model.f_pvalue
    
    return b, b_int, R2, F, p


# VIP: PLS VIP (variable importance in the projection)

import numpy as np

def vip(X,W,SSY):

    '''
    
    VIP: PLS variable importance in the projection
    [VIP] = VIP(X,nb)
    
    INPUT
    X [n x k] <numpy.ndarray>
        X data used in PLS
        n samples
        k variables
    W [k x nbPC] <numpy.ndarray>
        PLS weights
        k variables
        nbPC PLS components
    SSY [nbPC x 1] <numpy.ndarray>
        PLS sum of squares of Y
        nbPC PLS components      
    
    
    OUTPUT
    VIP [k x 1] 
        PLS VIP values
   
    
    '''
    
    
    VIP = []
    
    for i in range(X.shape[1]):
        wi = W[i,:]
        NUM = 0
         
        for j in range(W.shape[1]):
            wij = wi[j]
            num = wij**2 * SSY[j]
            NUM = NUM + num
    
        vip = X.shape[1]*NUM/np.sum(SSY)
        VIP.append(vip)
        
    VIP = np.array(VIP)



    return VIP